#
# snake/treedb.py - functions for manipulating the tree database
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#

import sys, os, time, tempfile
import snake.uri
import snake.tree
from snake.config import getoption
from ConfigParser import SafeConfigParser
import logging

def _loadtrees(dir=None,check=True):
    '''Loads all the treeinfo files in the given dir.
    Returns a list of Tree objects.
    If check is True, any treeinfo file lacking URIs will be removed from the
    treedb.'''
    if dir is None:
        dir = getoption("treedb_dir")
    treelist = list()
    for name in os.listdir(dir):
        f = os.path.join(dir,name)
        t = snake.tree.Tree()
        t.parsetreeinfo(f)
        if check and not (hasattr(t,'uris') and t.uris):
            removetree(t.id)
        treelist.append(t)
    return treelist

def _writetree(tree):
    '''Write the given Tree object back to the tree database'''
    # FIXME: safen tree.id
    treefile = os.path.join(getoption("treedb_dir"),"%s.treeinfo" % tree.id)
    fp = open(treefile,"w")
    tree.config.write(fp)
    fp.close()

def id_to_tree(id):
    treefile = os.path.join(getoption("treedb_dir"),"%s.treeinfo" % id)
    if not os.path.exists(treefile):
        return None
    try:
        t = snake.tree.Tree()
        t.parsetreeinfo(treefile)
    except ConfigParser.NoSectionError:
        # File exists but there's nobody home. Bluh.
        return None
    return t

def lookup_tree_by_uri(uri):
    '''Given a uri, search the local treedb for the Tree object containing it.
    Return a Tree object, or None on failure.'''
    treelist = _loadtrees()
    tree = None
    for t in treelist:
        if uri in t.uris:
            tree = t
            break
    return t

def add(uri):
    '''Add the given URI to the tree database'''
    # Get the new tree's info

    fetcher = snake.uri._fetcherForURI(uri)
    fetcher.prepareLocation(None)

#    t = snake.tree.Tree()
#    if fetcher.hasFile('.treeinfo'):
#        t.parsetreeinfo(fetcher.loadFile('.treeinfo', None))
#        logging.debug("treeinfo: %s, %s" % (t.id, t.name))
#    elif fetcher.hasFile('.discinfo'):
#        t.parsediscinfo(fetcher.loadFile('.discinfo', None))
#        logging.debug("discinfo: %s, %s" % (t.id, t.name))
#    else:
#        raise ValueError, "Unrecognized tree"

    t = snake.uri.uri_to_tree(uri, fetcher=fetcher)

    # open up db record
    c = SafeConfigParser()
    treefile = os.path.join(getoption("treedb_dir"),"%s.treeinfo" % t.id)
    if os.path.isfile(treefile):
        logging.debug("Loading existing treeinfo '%s'" % treefile)
        c.read(treefile)
        t.parsetreeinfo(treefile)

    # create a general section
    if not c.has_section("general"):
        c.add_section("general")
    for opt in ["id", "family", "timestamp", "version", "arch", "uris", "variant"]:
        c.has_option("general", opt) or c.set("general", opt, str(getattr(t, opt)))

    # create images sections
    for arch in t.images.keys():
        logging.debug("Adding section 'images-%s'" % arch)
        if not c.has_section("images-%s" % arch):
            c.add_section("images-%s" % arch)
            for file,paths in t.images[arch].items():
                logging.debug("paths = '%s'" % paths)
                if isinstance(paths,str):
                    paths = [paths]
                for path in paths:
                    logging.debug("hasFile('%s')" % path)
                    fetcher.hasFile(path) and c.set("images-%s" % arch, file, path)
        if len(c.options("images-%s" % arch)) == 0:
            c.remove_section("images-%s" % arch)
            t.images.pop(arch, True) # the True forces this to pass

    # FIXME - if no images are available ... this tree won't boot

    # add new uri
    if uri not in t.uris:
        t.uris.append(uri)
        c.set("general", "uris", "\n".join(t.uris))
        fp = open(treefile,"w")
        c.write(fp)
        fp.close()

    fetcher.cleanupLocation()
    return True

def listids(dir=None):
    '''Return a list of known tree IDs.'''
    if dir is None:
        dir = getoption("treedb_dir")
    idlist = list()
    for name in os.listdir(dir):
        (id,ext) = os.path.splitext(name)
        if ext != '.treeinfo':
            continue
        idlist.append(id)
    return idlist

def listtrees(**args):
    '''Return a list of Trees that match all the given args. Recognized args:
    id:     Tree id
    uri:    URI to a tree
    arch:   architecture
    family: family name (e.g. Fedora)
    ...and basically any other attribute of the Tree object.
    '''
    treelist = _loadtrees()
    resultlist = list()
    for t in treelist:
        ok = True
        for (k,v) in args.items():
            if (k == "uri" and v not in t.uris) or \
               (hasattr(t,k) and getattr(t,k) != v):
                ok = False
                break
        if ok:
            resultlist.append(t)
    return resultlist

def dicttrees(args={}):
    '''get a list of Tree objects (in dict format) that match the given args.
    If no args are given, will return a list of all trees.
    Common args:
    arch (e.g. i386, ppc, x86_64...)
    family (e.g. Fedora...)
    version (e.g. 6, 7, development...)
    '''
    treelist = listtrees(**args)
    for t in treelist:
        del t.config

    return [t.__dict__ for t in treelist]

    # TODO? return a sorted dictionary
    undecorated = [t.__dict__ for t in treelist]
    decorated = [(dict_['id'], dict_) for dict_ in undecorated]
    decorated.sort()
    return [dict_ for (key, dict_) in decorated]

def removeuri(uri):
    '''Remove the given URI from the tree database'''
    t = lookup_tree_by_uri(uri)
    if uri in t.uris:
        t.uris.remove(uri)
    # Tree has no remaining URIs - remove it.
    if not t.uris:
        return removetree(t.id)
    else:
        # FIXME this needs to happen in the tree object somehow
        t.config.set('general','uris',"\n".join(t.uris))
        _writetree(t)
        logging.debug("Removed uri '%s' from %s" % (uri, t.id))
    return True # XMLRPC methods must return a value

def removetree(id):
    '''Remove the given tree from the tree database'''
    # This is actually pretty easy.
    treefile = os.path.join(getoption("treedb_dir"),"%s.treeinfo" % id)
    if os.path.isfile(treefile):
        os.unlink(treefile)
        logging.debug("Removed tree '%s'" % treefile)
        return True # XMLRPC methods must return a value
    else:
        return False # XMLRPC methods must return a value
